#!/usr/bin/python

import socket, os, sys, struct, signal, errno
from struct import pack,unpack
from types import *

MSG_GET_PINFO = 1
MSG_PINFO = 2
MSG_QUIT = 3
MSG_GET_CWD = 4
MSG_CWD = 5
MSG_SIGTERM = 6
MSG_GET_ARGV = 7
MSG_ARGV = 8
MSG_GET_ENV = 9
MSG_ENV = 10
MSG_FORK_AND_RET = 11

class DeadSockError(RuntimeError) :
    pass

_lsock = None

def run_listener():
    global _lsock
    try:
        os.unlink("/tmp/initng-test.sock")
    except OSError:
        pass
    _lsock = socket.socket(socket.AF_UNIX,socket.SOCK_STREAM)
    _lsock.bind("/tmp/initng-test.sock")
    _lsock.listen(5)
    os.chmod("/tmp/initng-test.sock",0777)
    _lsock.settimeout(30.0) # seconds

def accept():
    try:
        s = _lsock.accept()[0]
    except socket.timeout:
        return None
    if s != None:
        return SlaveConn(s)

class SlaveConn:
    def __init__(self,sock):
        self._sock = sock
        self._q = [ ]
        self.settimeout(30.0)

    def settimeout(self,timeout):
        self._sock.settimeout(timeout)

    def get_msg(self):
        if len(self._q) == 0:
            return self._get_msg()
        else:
            return self._q.pop(0)

    def _get_msg_2(self,type):
        while True:
            msg = self._get_msg()
            if msg == None:
                break
            if msg[0] == type:
                break
            self._q.append(msg)
        return msg

    def _get_msg(self):
        return _get_msg(self._sock)
    
    def put_msg(self,msg,data=""):
        _put_msg(self._sock,msg,data)

    def get_pinfo(self):
        self.put_msg(MSG_GET_PINFO)
        msg = self._get_msg_2(MSG_PINFO)
        if msg == None: return None
        return unpack("!qqq",msg[1]) # pid, uid, gid

    def get_cwd(self):
        self.put_msg(MSG_GET_CWD)
        msg = self._get_msg_2(MSG_CWD)
        if msg == None: return None
        return msg[1]

    def get_argv(self):
        self.put_msg(MSG_GET_ARGV)
        msg = self._get_msg_2(MSG_ARGV)
        if msg == None: return None
        return msg[1]

    def get_env(self,key):
        self.put_msg(MSG_GET_ENV,key)
        while True:
            msg = self._get_msg_2(MSG_ENV)
            if msg == None:
                return None
            elif msg[1].startswith(key+"="):
                return msg[1][len(key)+1:]
            else:
                self._q.append(msg)
                
    def fork_and_ret(self):
        self.put_msg(MSG_FORK_AND_RET)

    def quit(self,ret=0):
        self.put_msg(MSG_QUIT,pack("!i",ret))
    
def _msg_get_pinfo(sock,data):
    _put_msg(sock,MSG_PINFO,pack("!qqq", os.getpid(),os.getuid(),os.getgid()))


def _msg_get_cwd(sock,data):
    _put_msg(sock,MSG_CWD,os.getcwd())

def _msg_quit(sock,data):
    code = 0
    try:
        code = unpack("!i",data)[0]
    except struct.error:
        sys.stderr.write("test_slave.py: couldn't get return code from MSG_QUIT\n")
    sys.stderr.write("test_slave.py: exiting with code %i\n" % code)
    sys.exit(code)

def _msg_get_argv(sock,data):
    odata = ""
    for arg in sys.argv:
        odata += arg + "\x00"
    _put_msg(sock, MSG_ARGV, odata)

def _msg_get_env(sock,data):
    if os.environ.has_key(data):
        _put_msg(sock, MSG_ENV, data + "=" + os.environ[data])
    else:
        _put_msg(sock, MSG_ENV, data + "=")

def _msg_fork_and_ret(sock,data):
    pid = os.fork()
    if pid == 0:
        pass # child
    else:
        os._exit(0) # parent

_handlers = { MSG_GET_PINFO: _msg_get_pinfo,
              MSG_GET_CWD: _msg_get_cwd,
              MSG_QUIT: _msg_quit,
              MSG_GET_ARGV: _msg_get_argv,
              MSG_GET_ENV: _msg_get_env,
              MSG_FORK_AND_RET: _msg_fork_and_ret
              }

def _put_msg(sock,msg,data=""):
    sock.sendall(pack("!ii",msg,len(data))+data)

def _get_msg(sock):
    try:
        h = sock.recv(struct.calcsize("!ii"))
    except socket.timeout:
        #sys.stderr.write("test_slave.py: timeout getting message header\n")
        return None
    except socket.error, (eno, strerr):
        if eno != errno.EINTR and eno != errno.EAGAIN:
            raise
        else:
            return None
    if len(h) == 0:
        raise DeadSockError
    if len(h) != struct.calcsize("!ii"):
        sys.stderr.write("test_slave.py: couldn't get message header\n")
        return None
    h = unpack("!ii",h)
    msg = h[0]
    if h[1] == 0:
        data = ""
    else:
        try:
            data = sock.recv(h[1])
        except socket.timeout:
            sys.stderr.write("test_slave.py: timeout getting message data\n")
            return None
    if len(data) != h[1]:
        sys.stderr.write("test_slave.py: couldn't get message payload\n")
        return None
    return (msg,data)

_sigterm_flg = False

def _sigterm_handler(signum, frame):
    global _sigterm_flg
    _sigterm_flg = True

def _main():
    global _sigterm_flg
    sys.stderr.write("test_slave.py: connecting to socket...\n")
    signal.signal(signal.SIGTERM, _sigterm_handler)
    s = socket.socket(socket.AF_UNIX,socket.SOCK_STREAM)
    s.connect("/tmp/initng-test.sock")
    sys.stderr.write("test_slave.py: connected!\n")
    s.settimeout(7.0)
    while True:
        try:
            msg = _get_msg(s)
        except socket.timeout:
            msg = None
        if(_sigterm_flg):
            print "Caught SIGTERM!"
            _sigterm_flg = False
            _put_msg(s,MSG_SIGTERM)
        if msg == None: continue
        print msg
        if _handlers.has_key(msg[0]):
            _handlers[msg[0]](s,msg[1])
        

if __name__=='__main__':
    _main()
